# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: disable=redefined-outer-name, no-member
""" Common routines, classes, fixtures for CC tests. """

import pytest
import numpy as np

from pybest.context import context
from pybest.io.iodata import IOData
from pybest.gbasis.gobasis import PYBEST_CHOLESKY_ENABLED
from pybest.gbasis import (
    get_gobasis,
    compute_kinetic,
    compute_nuclear,
    compute_eri,
    compute_cholesky_eri,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.geminals.rpccd import RpCCD
from pybest.linalg import (
    FourIndex,
    DenseLinalgFactory,
    CholeskyLinalgFactory,
    DenseTwoIndex,
)
from pybest.scf import AufbauOccModel
from pybest.wrappers.hf import RHF

# Define a set of linear algebra factories to be tested
if PYBEST_CHOLESKY_ENABLED:
    LINALG = [DenseLinalgFactory, CholeskyLinalgFactory]
else:
    LINALG = [DenseLinalgFactory]


class Model:
    """ Handles all common pre-CC operations. """

    def __init__(self, path_xyz, npairs, basis, read=True):
        self.occ_model = AufbauOccModel(npairs)
        self.init_input(context.get_fn(path_xyz), basis)
        name = path_xyz.split(".")[0]
        rhf = RHF(self.lf, self.occ_model)
        self.rhf = rhf(*self.ints(), self.olp, self.orb_a)
        if read:
            # Read pCCD orbitals and compute other pCCD-related quantities
            data = IOData.from_file(context.get_fn(name + "_pccd.molden"))
            pccd = RpCCD(self.lf, self.occ_model)
            self.pccd = pccd(*self.ints(), self.olp, data.orb_a, e_core=0.0)

    def init_input(self, xyz_file, basis):
        " Construct basis object and linear algebra factory, compute integrals."
        obasis = get_gobasis(basis, xyz_file, print_basis=False)
        self.lf = DenseLinalgFactory(obasis.nbasis)
        self.olp = compute_overlap(obasis)
        self.orb_a = self.lf.create_orbital(obasis.nbasis)
        # Integrals
        self.kin = compute_kinetic(obasis)
        self.nuc = compute_nuclear(obasis)
        self.eri = compute_eri(obasis)
        self.cholesky_eri = None
        if PYBEST_CHOLESKY_ENABLED:
            self.cholesky_eri = compute_cholesky_eri(obasis, threshold=1e-9)
        self.ext = compute_nuclear_repulsion(obasis)

    def ints(self, linalg_factory=DenseLinalgFactory):
        " Returns list of one-body and two-body integrals."
        if linalg_factory is DenseLinalgFactory:
            return [self.kin, self.nuc, self.eri]
        return [self.kin, self.nuc, self.cholesky_eri]


def check_fock_in_cache(cache, labels, nocc=5, nvirt=8):
    "Checks if labels correspond to Fock matrix blocks in cache."
    dim = {"o": nocc, "v": nvirt}
    for label in labels:
        msg = f"{label} block in cc.hamiltonian: \n"
        matrix = cache.load(label)
        assert isinstance(matrix, DenseTwoIndex), msg + "incorrect type"
        assert matrix.shape[0] == dim[label[-2]], msg + "incorrect size"
        assert matrix.shape[1] == dim[label[-1]], msg + "incorrect size"
        # occupied-virtual block for RHF orbitals is zeros by nature
        if not label == "fock_ov":
            is_zeros = np.allclose(matrix.array, np.zeros(matrix.shape))
            assert not is_zeros, msg + " is filled with zeros!"


def check_eri_in_cache(cache, labels, nocc=5, nvirt=8):
    "Checks if labels correspond to 2-body CC Hamiltonian blocks in cache."
    dim = {"o": nocc, "v": nvirt}
    for label in labels:
        msg = f"Checking {label} block in cc.hamiltonian...\n"
        matrix = cache.load(label)
        assert isinstance(matrix, FourIndex), msg + "wrong type!"
        for i in range(4):
            assert matrix.shape[i] == dim[label[i - 4]], msg + "incorrect size"
        assert not np.allclose(matrix.array, np.zeros(matrix.array.shape))
        assert not np.isnan(matrix.array).any()


@pytest.fixture
def h2_molecule():
    "Returns RHF/pCCD results and reference CC energies."
    model = Model("test/h2.xyz", 1, "cc-pvdz")
    reference_correlation_energy = {
        "RCCD": -0.034582180698988885,
        "RCCSD": -0.034709514282981926,
        "RfpCCD": -0.034840613788246884,
        "RfpCCSD": -0.03484061735867886,
    }
    return {"model": model, "energies": reference_correlation_energy}


@pytest.fixture
def hf_molecule():
    "Returns RHF/pCCD results and reference CC energies."
    model = Model("test/hf.xyz", 5, "3-21g")
    reference_correlation_energy = {
        "RCCD": -0.12911913585251084,
        "RCCSD": -0.12987666859843192,
        "RfpCCD": -0.13459134995132008,
        "RfpCCSD": -0.13465477029776618,
    }
    return {"model": model, "energies": reference_correlation_energy}


@pytest.fixture
def h2o_molecule():
    "Returns RHF/pCCD results and reference CC energies."
    model = Model("test/h2o.xyz", 5, "3-21g")
    reference_correlation_energy = {
        "RCCD": -0.1499368041545847,
        "RCCSD": -0.15213686280456384,
        "RfpCCD": -0.16072464912359646,
        "RfpCCSD": -0.16094396980907438,
    }
    return {"model": model, "energies": reference_correlation_energy}
