# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""
   Variables used in this module:
    :ncore:     number of frozen core orbitals
    :nocc:      number of occupied orbitals in the principle configuration
    :nacto:     number of active occupied orbitals in the principle configuration
    :nvirt:     number of virtual orbitals in the principle configuration
    :nactv:     number of active virtual orbitals in the principle configuration
    :nbasis:    total number of basis functions
    :nact:      total number of active orbitals (nacto+nactv)
    :e_ci:      eigenvalues of CI Hamiltonian (IOData container attribute)
    :civ:       eigenvectors of CI Hamiltonian (IOData container attribute)

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
"""
import numpy as np

from pybest.ci.rci_base import RCI
from pybest.ci.rci_utils import display, set_dimension
from pybest.log import log
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.auxmat import get_fock_matrix
from pybest.exceptions import NonEmptyData
from pybest.utils import check_options


class RCID(RCI):
    """Restricted Configuration Interaction Doubles module
    for Slater Determinant (SD) and Configuration State Function (CSF) basis.
    """

    long_name = "Restricted Configuration Interaction Doubles"
    acronym = "CID"
    reference = "any single-reference wavefunction"

    @RCI.dimension.setter
    def dimension(self, new=None):
        if new is not None:
            self._dimension = set_dimension(new, self.nacto, self.nactv)
        else:
            self._dimension = set_dimension(
                self.acronym, self.nacto, self.nactv
            )

    @RCI.nroot.setter
    def nroot(self, new):
        self._nroot = new

    @RCI.size_consistency_correction.setter
    def size_consistency_correction(self, new):
        self._size_consistency_correction = new

    @RCI.threshold_c_0.setter
    def threshold_c_0(self, new):
        self._threshold_c_0 = new

    def build_subspace_hamiltonian(self, bvector, hamiltonian, *args):
        """
        Used by davidson module to construct subspace Hamiltonian.
        Calls two subfunctions dedicated for either SD
        (build_subspace_hamiltonian_sd) or CSF basis.

        **Arguments:**

        bvector:
            (OneIndex object) contains current approximation to CI coefficients.

        hamiltonian:
            (OneIndex object) used by davidson module and contains
            diagonal approximation to the full matrix.
        """
        if self.csf:
            raise NotImplementedError
        return self.build_subspace_hamiltonian_sd(bvector)

    def build_subspace_hamiltonian_sd(self, bvector):
        """
        Constructing Hamiltonian subspace for Slater Determinant basis

        **Arguments:**

        bvector:
            (OneIndex object) contains current approximation to CI coefficients
            in the SD basis.
        """
        #
        # Integrals
        #
        fock = self.load_aux_tensor("fock")
        govvo = self.load_aux_tensor("govvo")
        govov = self.load_aux_tensor("govov")
        goooo = self.load_aux_tensor("goooo")
        eri = self.load_aux_tensor("eri")

        #
        # Ranges
        #
        oo = self.get_range("oo")
        vv = self.get_range("vv")
        vvvv = self.get_range("vvvv")
        #
        # TEMP OBJ.
        #
        f_oo = fock.copy(**oo)
        f_vv = fock.copy(**vv)
        #
        # local variables
        #
        nacto = self.nacto
        nactv = self.nactv
        tco = {"select": self.tco}
        end_ab = self.nacto * self.nacto * self.nactv * self.nactv + 1
        size_aa = nacto * (nacto - 1) * nactv * (nactv - 1) // 4
        #
        # Bvectors
        #
        b_dab = self.denself.create_four_index(
            self.nacto, self.nactv, self.nacto, self.nactv
        )
        #
        # Bvectors assignment
        #
        b_dab.assign(bvector, begin4=1, end4=end_ab)
        #
        # Sigma vector
        #
        sigma_d = self.denself.create_four_index(
            self.nacto, self.nactv, self.nacto, self.nactv
        )

        #
        # block alpha-beta (Opposite spins (iaJB))
        #
        # 1) <Ja|Bi> * r_0
        govvo.slice("abcd->dbac", out=sigma_d, factor=bvector.get_element(0))

        # 2) <kD|cL> r_kcLD
        value_0 = b_dab.contract("abcd,adbc", govvo)

        # 3) -f_ki r_kajb
        b_dab.contract("abcd,ae->ebcd", f_oo, out=sigma_d, factor=-1.0, **tco)

        # 4) -f_lj r_ialb
        b_dab.contract("abcd,ce->abed", f_oo, out=sigma_d, factor=-1.0, **tco)

        # 5) +f_ac r_icjb
        b_dab.contract("abcd,eb->aecd", f_vv, out=sigma_d, **tco)

        # 6) +f_bd r_iajd
        b_dab.contract("abcd,ed->abce", f_vv, out=sigma_d, **tco)

        # 7) +<ka|ci> r_kcJB
        b_dab.contract("abcd,aebf->fecd", govvo, out=sigma_d, **tco)

        # 8) -<ka|ic> r_kcJB
        b_dab.contract(
            "abcd,aefb->fecd", govov, out=sigma_d, factor=-1.0, **tco
        )

        # 9) +<LB|DJ> r_iaLD
        b_dab.contract("abcd,cedf->abfe", govvo, out=sigma_d, **tco)

        # 10) -<LB|JD> r_iaLD
        b_dab.contract(
            "abcd,cefd->abfe", govov, out=sigma_d, factor=-1.0, **tco
        )

        # 11) <aB|cD> r_icJD
        eri.contract("abcd,ecfd->eafb", b_dab, out=sigma_d, **vvvv)

        # 12) <kl|ij> r_kaLB
        b_dab.contract("abcd,acef->ebfd", goooo, out=sigma_d, **tco)

        # 13) -<kB|iD> r_kaJD
        b_dab.contract(
            "abcd,aefd->fbce", govov, out=sigma_d, factor=-1.0, **tco
        )

        # 14) -<La|Jc> r_icLB
        b_dab.contract(
            "abcd,cefb->aefd", govov, out=sigma_d, factor=-1.0, **tco
        )
        #
        # Assign to sigma vector
        #
        sigma = self.lf.create_one_index(self.dimension)
        sigma.assign(sigma_d.array.ravel(), begin0=1, end0=end_ab)

        if self.nacto > 1:
            #
            # block alpha-alpha (iajb) / beta-beta (IAJB)
            #
            b_daa = self.denself.create_four_index(
                self.nacto, self.nactv, self.nacto, self.nactv
            )
            end_aa = nacto * (nacto - 1) * nactv * (nactv - 1) // 4 + end_ab
            for shift in [0, size_aa]:
                sigma_d.clear()
                b_daa.clear()
                b_daa.assign(
                    bvector,
                    ind=self.get_index_of_mask(),
                    begin4=end_ab + shift,
                    end4=end_aa + shift,
                )
                # create tmp object to account for symmetry
                tmp = b_daa.copy()
                b_daa.iadd_transpose((2, 1, 0, 3), other=tmp, factor=-1.0)
                b_daa.iadd_transpose((0, 3, 2, 1), other=tmp, factor=-1.0)
                b_daa.iadd_transpose((2, 3, 0, 1), other=tmp)
                del tmp
                # 1) <kl||cd> r_kcld (<kd|cl> - <kc|dl>)
                value_0 += b_daa.contract("abcd,adbc", govvo, factor=0.25)
                value_0 += b_daa.contract("abcd,abdc", govvo, factor=-0.25)
                # 2) <ab||ij> * r_0 (<ib|aj> - <ia|bj>)
                govvo.contract(
                    "abcd->acdb", out=sigma_d, factor=bvector.get_element(0)
                )
                govvo.contract(
                    "abcd->abdc", out=sigma_d, factor=-bvector.get_element(0)
                )
                # 3) -f_ki r_kajb
                b_daa.contract("abcd,ae->ebcd", f_oo, sigma_d, factor=-1.0)
                # 4) -f_lj r_ialb
                b_daa.contract("abcd,ce->abed", f_oo, sigma_d, factor=-1.0)
                # 5) f_ad r_idjb
                b_daa.contract("abcd,eb->aecd", f_vv, sigma_d)
                # 6) f_bd r_iajd
                b_daa.contract("abcd,ed->abce", f_vv, sigma_d)
                # 7) <ab||cd> r_icjd
                eri.contract(
                    "abcd,ecfd->eafb", b_daa, sigma_d, factor=0.5, **vvvv
                )
                eri.contract(
                    "abcd,edfc->eafb", b_daa, sigma_d, factor=-0.5, **vvvv
                )
                # 8) <ij||kl> r_kalb
                goooo.contract("abcd,cedf->aebf", b_daa, sigma_d, factor=0.5)
                goooo.contract("abcd,decf->aebf", b_daa, sigma_d, factor=-0.5)
                # 9) <lb||dj> r_iald (<lb|dj> - <lb|jd>)
                b_daa.contract("abcd,cedf->abfe", govvo, sigma_d)
                b_daa.contract("abcd,cefd->abfe", govov, sigma_d, factor=-1.0)
                # 10) -<lb||di> r_jald (-<lb|di> + <lb|id>)
                b_daa.contract("abcd,cedf->fbae", govvo, sigma_d, factor=-1.0)
                b_daa.contract("abcd,cefd->fbae", govov, sigma_d)
                # 11) <ka||ci> r_kcjb (+<ka|ci> - <ka|ic>)
                b_daa.contract("abcd,aebf->fecd", govvo, sigma_d)
                b_daa.contract("abcd,aefb->fecd", govov, sigma_d, factor=-1.0)
                # 12) -<ka||cj> r_kcib (-<ka|cj> + <ka|jc>))
                b_daa.contract("abcd,aebf->cefd", govvo, sigma_d, factor=-1.0)
                b_daa.contract("abcd,aefb->cefd", govov, sigma_d)
                #
                # Coupling terms between same spin and opposite spin
                #
                # 13) <iajb|H|kcLD> and <iajb|H|KCld>
                #
                if shift == 0:
                    # a) -<La|Dj>(r_ibLD) * bvector
                    b_dab.contract(
                        "abcd,cedf->aefb", govvo, sigma_d, factor=-1.0
                    )
                    # b) +<La|Di>(r_jbLD) * bvector
                    b_dab.contract(
                        "abcd,cedf->feab", govvo, sigma_d, factor=1.0
                    )
                    # c) <Lb|Dj>(r_iaLD) * bvector
                    b_dab.contract(
                        "abcd,cedf->abfe", govvo, sigma_d, factor=1.0
                    )
                    # d) -<Lb|Di>(r_jaLD) * bvector
                    b_dab.contract(
                        "abcd,cedf->fbae", govvo, sigma_d, factor=-1.0
                    )
                else:
                    # a) -<Ka|Cj>(r_KCib) * bvector
                    b_dab.contract(
                        "abcd,aebf->cefd", govvo, sigma_d, factor=-1.0
                    )
                    # b) +<Ka|Ci>(r_KCjb) * bvector
                    b_dab.contract(
                        "abcd,aebf->fecd", govvo, sigma_d, factor=1.0
                    )
                    # c) <Kb|Cj>(r_KCia) * bvector
                    b_dab.contract(
                        "abcd,aebf->cdfe", govvo, sigma_d, factor=1.0
                    )
                    # d) -<Kb|Ci>(r_KCja) * bvector
                    b_dab.contract(
                        "abcd,aebf->fdce", govvo, sigma_d, factor=-1.0
                    )
                sigma.assign(
                    sigma_d.array[self.get_mask()],
                    begin0=end_ab + shift,
                    end0=end_aa + shift,
                )
                sigma_d.clear()
                #
                # Coupling terms between opposite spin and same spin
                #
                # 14) <iaJB|H|kcld> and <iaJB|H|KCLD>
                #
                if shift == 0:
                    b_daa.contract("abcd,cedf->abfe", govvo, sigma_d, **tco)
                else:
                    b_daa.contract("abcd,aebf->fecd", govvo, sigma_d, **tco)
                sigma.iadd(sigma_d.array.ravel(), begin0=1, end0=end_ab)

        sigma.set_element(0, value_0)
        return sigma

    def compute_h_diag(self, *arg):
        """Used by davidson module for pre-conditioning.

        **Returns:**
            (OneIndex object) contains guess vector to davidson diagonalization.

        """
        if self.csf:
            raise NotImplementedError
        return self.h_diag_sd()

    def h_diag_sd(self):
        """Calculating the guess vector for SD basis.

        hdiag:
            (OneIndex object) contains guess vector for SD basis.
        """
        #
        # Auxiliary objects
        #
        fock = self.load_aux_tensor("fock")
        govvo = self.load_aux_tensor("govvo")
        govov = self.load_aux_tensor("govov")
        goooo = self.load_aux_tensor("goooo")
        eri = self.load_aux_tensor("eri")

        #
        # local variables
        #
        hdiag = self.lf.create_one_index(self.dimension)

        #
        # Ranges
        #
        vvvv = self.get_range("vvvv")
        end_ab = self.nacto * self.nacto * self.nactv * self.nactv + 1
        if self.nacto > 1:
            end_aa = (
                self.nacto
                * (self.nacto - 1)
                * self.nactv
                * (self.nactv - 1)
                // 4
                + end_ab
            )

        #
        # Intermediates
        #
        fii = self.lf.create_one_index(self.nacto)
        fock.copy_diagonal(out=fii, end=self.nacto)

        faa = self.lf.create_one_index(self.nactv)
        fock.copy_diagonal(out=faa, begin=self.nacto, end=self.nact)

        g_ijij = self.lf.create_two_index(self.nacto, self.nacto)
        goooo.slice("abab->ab", g_ijij)
        g_ijji = self.lf.create_two_index(self.nacto, self.nacto)
        goooo.slice("abba->ab", g_ijji, factor=-1.0)

        g_abab = self.lf.create_two_index(self.nactv, self.nactv)
        eri.slice("abab->ab", g_abab, **vvvv)
        g_abba = self.lf.create_two_index(self.nactv, self.nactv)
        eri.slice("abba->ab", g_abba, factor=-1.0, **vvvv)

        g_iaia = self.lf.create_two_index(self.nacto, self.nactv)
        govov.slice("abab->ab", g_iaia)
        g_iaai = self.lf.create_two_index(self.nacto, self.nactv)
        govvo.slice("abba->ab", g_iaai, factor=-1.0)

        r_iajb = self.denself.create_four_index(
            self.nacto, self.nactv, self.nacto, self.nactv
        )

        #
        # -f_ii - f_jj
        #
        r_iajb.iadd_expand_one_to_four("0", fii, factor=-1.0)
        r_iajb.iadd_expand_one_to_four("2", fii, factor=-1.0)

        #
        # +f_aa + f_bb
        #
        r_iajb.iadd_expand_one_to_four("1", faa)
        r_iajb.iadd_expand_one_to_four("3", faa)

        #
        # <ab|ab>
        #
        r_iajb.iadd_expand_two_to_four("bd->abcd", g_abab)
        #
        # <ij|ij>
        #
        r_iajb.iadd_expand_two_to_four("ac->abcd", g_ijij)

        #
        # - <jb||jb>
        #
        r_iajb.iadd_expand_two_to_four("cd->abcd", g_iaia, factor=-1.0)
        r_iajb.iadd_expand_two_to_four("cd->abcd", g_iaai, factor=-1.0)
        #
        # - <ib|ib>
        #
        r_iajb.iadd_expand_two_to_four("ad->abcd", g_iaia, factor=-1.0)
        #
        # - <ja|ja>
        #
        r_iajb.iadd_expand_two_to_four("cb->abcd", g_iaia, factor=-1.0)
        #
        # - <ia||ia>
        #
        r_iajb.iadd_expand_two_to_four("ab->abcd", g_iaia, factor=-1.0)
        r_iajb.iadd_expand_two_to_four("ab->abcd", g_iaai, factor=-1.0)
        #
        # Assign opposite spin block
        #
        hdiag.assign(r_iajb.array.ravel(), begin0=1, end0=end_ab)

        #
        # Assign same spin block
        #
        if self.nacto > 1:
            #
            # same spin terms
            #
            # <ab|ba>
            #
            r_iajb.iadd_expand_two_to_four("bd->abcd", g_abba)
            #
            # <ij|ji>
            #
            r_iajb.iadd_expand_two_to_four("ac->abcd", g_ijji)
            hdiag.assign(
                r_iajb.array[self.get_mask()], begin0=end_ab, end0=end_aa
            )
            hdiag.assign(r_iajb.array[self.get_mask()], begin0=end_aa)

        return hdiag

    def calculate_aux_tensors(self, mo1, mo2):
        """Compute auxiliary matrices.

        **Arguments:**

        mo1, mo2
            One- and two-electron integrals (some Hamiltonian matrix
            elements) in the MO basis.
        """
        self.clear_aux_tensors()
        #
        # Get ranges
        #
        ovvo = self.get_range("ovvo")
        ovov = self.get_range("ovov")
        oooo = self.get_range("oooo")

        # 1) Fock matrix: fpq
        fock = self.init_aux_tensors("fock", self.nact, self.nact)
        get_fock_matrix(fock, mo1, mo2, self.nacto)

        # 2) temporary matrix: govvo
        govvo = self.init_aux_tensors(
            "govvo", self.nacto, self.nactv, self.nactv, self.nacto
        )
        mo2.slice("abcd->abcd", govvo, **ovvo)

        # 3) temporary matrix: govov
        govov = self.init_aux_tensors(
            "govov", self.nacto, self.nactv, self.nacto, self.nactv
        )
        mo2.slice("abcd->abcd", govov, **ovov)

        # 4) temporary matrix: goooo
        goooo = self.init_aux_tensors(
            "goooo", self.nacto, self.nacto, self.nacto, self.nacto
        )
        mo2.slice("abcd->abcd", goooo, **oooo)

        # 5) temporary matrix: gppqq
        gppqq = self.init_aux_tensors("gppqq", self.nact, self.nact)
        mo2.slice("aabb->ab", gppqq)

        # 6) temporary matrix: eri = electron repulsion integrals
        if isinstance(self.lf, CholeskyLinalgFactory):
            eri = self.init_aux_tensors(
                "eri",
                self.nact,
                nvec=mo2.nvec,
            )
            eri.assign(
                mo2,
            )
        else:
            eri = self.init_aux_tensors("eri", self.nact)
            eri.assign(
                mo2,
            )

    def init_aux_tensors(
        self, select, dim0=None, dim1=None, dim2=None, dim3=None, nvec=None
    ):
        # pylint: disable=too-many-arguments
        """Initialize auxiliary matrices.

        **Arguments:**

        select
             (string) label of the auxiliary matrix.

        **Optional Arguments:**

        dim0, dim1, dim2, dim3
             The size of the auxiliary matrix in each dimension. The number of given
             arguments not equal to None determines the order of the matrix.
        """
        check_options(
            "select",
            select,
            "fock",
            "goooo",
            "govov",
            "govvo",
            "eri",
            "gppqq",
        )

        if select in ["fock", "gppqq"]:
            alloc_size = (self.lf.create_two_index, dim0, dim1)
        elif select in ["govvo", "govov", "goooo"]:
            alloc_size = (
                self.denself.create_four_index,
                dim0,
                dim1,
                dim2,
                dim3,
            )
        elif select in ["eri"]:
            alloc_size = (self.lf.create_four_index, dim0, nvec)

        matrix, new = self._cache.load(
            select,
            alloc=alloc_size,
            tags="m",
        )
        if not new:
            raise NonEmptyData(
                f"The matrix {select} already exists. Call clear prior "
                f"to updating the wfn."
            )
        return matrix

    def printer(self):
        """Printing the results."""
        #
        # Local variables
        #
        evals = self.checkpoint["e_ci"]
        evecs = self.checkpoint["civ"]
        threshold = self.threshold
        alpha = "\u03B1"
        beta = "\u03B2"
        data = {
            "spin_block_aa",
            "spin_block_bb",
            "spin_block_ab",
            "c_ab",
            "c_aa",
            "c_bb",
        }
        #
        # Printing
        #
        log.hline("*")
        log(f"RESULTS OF {self.acronym}")
        log.hline("*")
        log(
            f"{'Root':>5s} {'Exc.Energy[au]':>16s} {'Tot.Energy[au]':>17s}"
            f" {'Weight(d)':>15s}"
        )
        log.hline("_")
        for ind, val in enumerate(evals):
            data = {key: list() for key in data}
            evecsj = evecs[:, ind]
            n_c = np.where(abs(evecsj) > threshold)[0]
            log(
                f"{ind:>3d} {val:> 17.6e} {self.e_ref+val:> 16.8f}"
                f"{np.dot(evecsj[1:], evecsj[1:]):> 17.6f}"
            )
            log(" ")
            for ind2 in n_c:
                if ind2 == 0:
                    if self.size_consistency_correction:
                        self.rci_corrections.printer()
                    log(f"{'Reference state    C_0:':>31s}")
                    log(f"{evecsj[ind2]:> 34.5f}")
                else:
                    self.collect_data(ind2, data, evecsj)

            if self.csf:
                raise NotImplementedError
            if (
                len(data["spin_block_aa"]) > 0
                or len(data["spin_block_ab"]) > 0
            ):
                log(" ")
                log(f"{'(i->a   j->b):':>22s} {'C_iajb:':>11s}")
            display(data, "ab", alpha, beta)
            display(data, "aa", alpha, alpha)
            display(data, "bb", beta, beta)

            log.hline("-")
            log.hline("*")

    def collect_data(self, index, data, evecsj):
        """Collect the data and prepare them for printing:

        **Arguments:**

        *index:
            (int) Number indicating the SD contribution in the CI solution.

        *data:
            (dictionary) Contains two types of data:
            indices (spin_block_[aa, bb or ab]) of the proper SD spin block
            contributions and the corresponding coefficients (c_[aa, bb or ab]).

        *evecsj:
            Eigenvectors of CI Hamiltonian (without the reference
            state contribution).
        """
        end_ab = self.nacto * self.nacto * self.nactv * self.nactv
        end_aa = end_ab + ((self.nacto * (self.nacto - 1)) // 2) * (
            (self.nactv * (self.nactv - 1)) // 2
        )

        i, a, j, b = self.get_index_d(index - 1)
        i, a, j, b = (
            i + self.ncore,
            a + self.ncore,
            j + self.ncore,
            b + self.ncore,
        )
        if index < end_ab:
            data["spin_block_ab"].append([i, a, j, b])
            data["c_ab"].append(evecsj[index])
        elif index < end_aa:
            data["spin_block_aa"].append([i, a, j, b])
            data["c_aa"].append(evecsj[index])
        else:
            data["spin_block_bb"].append([i, a, j, b])
            data["c_bb"].append(evecsj[index])

        return data

    def get_mask(self):
        """The function returns a 4-dimensional boolean np.array. True values
        are assigned to all non-redundant and symmetry-unique elements of the
        CI coefficient tensor for double excitations."""
        mask = np.zeros(
            (self.nacto, self.nactv, self.nacto, self.nactv), dtype=bool
        )
        indices_o = np.triu_indices(self.nacto, 1)
        indices_v = np.triu_indices(self.nactv, 1)

        for i, j in zip(indices_o[0], indices_o[1]):
            mask[i, indices_v[0], j, indices_v[1]] = True
        return mask

    def get_index_of_mask(self):
        """Get the indices where the True values are assigned."""
        mask = self.get_mask()
        indices = np.where(mask)
        return indices

    def get_index_d(self, index):
        """Get the unique indices of some doubly excited SD. Returns the set of
        active orbital indices without adding shift related to the frozen core
        orbitals.

        **Arguments:**

        *index:
            (int) The number that indicates doubly excited SD which
            is contributing in the CI solution.
        """
        # Case 1) C_iaJB:
        end_ab = self.nacto * self.nacto * self.nactv * self.nactv
        # Case 2) C_iajb:
        end_aa = ((self.nacto * (self.nacto - 1)) // 2) * (
            (self.nactv * (self.nactv - 1)) // 2
        )
        # 1) we store all i,a,j,b and we can simply use np's unravel_index build-in function
        if index < end_ab:
            (i, a, j, b) = np.unravel_index(
                index, (self.nacto, self.nactv, self.nacto, self.nactv)
            )
            i, a, j, b = (
                i + 1,
                a + 1 + self.nacto,
                j + 1,
                b + 1 + self.nacto,
            )
            return i, a, j, b
        # 2) This is more complicated, one possible way is to use a mask function:
        index = index - end_ab
        # 3) This is the same as case 2) but index has to be shifted by end_aa as well
        if index > end_aa:
            index = index - end_aa
        ind = np.where(self.get_mask())
        i, a, j, b = (
            ind[0][index],
            ind[1][index],
            ind[2][index],
            ind[3][index],
        )
        return (
            i + 1,
            a + 1 + self.nacto,
            j + 1,
            b + 1 + self.nacto,
        )

    def calculate_exact_hamiltonian(self):
        """Calculate the exact hamiltonian of CID model."""
        raise NotImplementedError
