# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
"""
   Variables used in this module:
    :ncore:     number of frozen core orbitals
    :nocc:      number of occupied orbitals in the principle configuration
    :nacto:     number of active occupied orbitals in the principle configuration
    :nvirt:     number of virtual orbitals in the principle configuration
    :nactv:     number of active virtual orbitals in the principle configuration
    :nbasis:    total number of basis functions
    :nact:      total number of active orbitals (nacto+nactv)
    :e_ci:      eigenvalues of CI Hamiltonian (IOData container attribute)
    :civ:       eigenvectors of CI Hamiltonian (IOData container attribute)

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
"""
import numpy as np

from pybest.ci.rci_base import RCI
from pybest.ci.rcid import RCID
from pybest.ci.rcis import RCIS
from pybest.ci.rci_utils import display, set_dimension
from pybest.log import log
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.auxmat import get_fock_matrix
from pybest.exceptions import NonEmptyData
from pybest.utils import check_options


class RCISD(RCID, RCIS, RCI):
    """Restricted Configuration Interaction Singles and Doubles module
    for Slater Determinant(SD) and Configuration State Function(CSF) basis.
    """

    long_name = "Restricted Configuration Interaction Singles and Doubles"
    acronym = "CISD"
    reference = "any single-reference wavefunction"

    @RCI.dimension.setter
    def dimension(self, new=None):
        if new is not None:
            self._dimension = set_dimension(new, self.nacto, self.nactv)
        else:
            self._dimension = set_dimension(
                self.acronym, self.nacto, self.nactv
            )

    @RCI.nroot.setter
    def nroot(self, new):
        self._nroot = new

    @RCI.size_consistency_correction.setter
    def size_consistency_correction(self, new):
        self._size_consistency_correction = new

    @RCI.threshold_c_0.setter
    def threshold_c_0(self, new):
        self._threshold_c_0 = new

    def build_subspace_hamiltonian(self, bvector, hamiltonian, *args):
        """
        Used by Davidson module to construct subspace Hamiltonian.
        Calls two subfunctions dedicated for either SD or CSF basis.
        Hamiltonian is prepared by mixing the contributions from CIS and CID
        and adding the couplings between Singles and Doubles blocks.

        **Arguments:**

        bvector:
            (OneIndex object) contains current approximation to CI coefficients.

        hamiltonian:
            (OneIndex object) used by davidson module and contains
            diagonal approximation to the full matrix.
        """
        if self.csf:
            raise NotImplementedError
        #
        # Integrals
        #
        fock = self.load_aux_tensor("fock")
        gooov = self.load_aux_tensor("gooov")
        eri = self.load_aux_tensor("eri")
        #
        # Ranges
        #
        ov = self.get_range("ov")
        ovvv = self.get_range("ovvv")
        #
        # Local variables
        #
        fia = fock.copy(**ov)
        nacto = self.nacto
        nactv = self.nactv
        end_ia = nacto * nactv + 1
        end_jb = end_ia + nacto * nactv
        end_ab = end_jb + nacto ** 2 * nactv ** 2
        #
        # Sigma Vectors
        #
        sigma = self.lf.create_one_index(self.dimension)
        sigma_s = self.denself.create_two_index(self.nacto, self.nactv)
        #
        # Bvectors
        #
        b_sa = self.lf.create_two_index(self.nacto, self.nactv)
        b_dab = self.denself.create_four_index(
            self.nacto, self.nactv, self.nacto, self.nactv
        )
        b_daa = self.denself.create_four_index(
            self.nacto, self.nactv, self.nacto, self.nactv
        )
        #
        # Bvectors assignment
        #
        b_sa.assign(bvector, begin2=1, end2=end_ia)
        b_dab.assign(bvector, begin4=end_jb, end4=end_ab)

        #
        # Assign Singles and Doubles block
        #
        singles = self.build_subspace_hamiltonian_singles(
            bvector.copy(begin=0, end=end_jb), hamiltonian
        )
        doubles = self.build_subspace_hamiltonian_doubles(bvector)

        sigma.set_element(0, singles.get_element(0) + doubles.get_element(0))
        sigma.assign(singles.array[1:], begin0=1, end0=end_jb)
        sigma.assign(doubles.array[1:], begin0=end_jb)

        del singles, doubles
        self.dimension = "CISD"

        #
        # Calculating Singles-Doubles Coupling terms.
        #
        for shift in [0, end_ia - 1]:
            sigma_d = self.denself.create_four_index(
                self.nacto, self.nactv, self.nacto, self.nactv
            )
            sigma_s.clear()
            b_sa.clear()
            b_sa.assign(bvector, begin2=1 + shift, end2=end_ia + shift)

            if shift == 0:
                # 1a)
                # block alpha: f_kc r_iakc
                b_dab.contract("abcd,cd->ab", fia, out=sigma_s, factor=1.0)
                # 2a)
                # block alpha: <la|dc> r_icld
                eri.contract(
                    "abcd,edac->eb", b_dab, out=sigma_s, factor=1.0, **ovvv
                )
                # 3a)
                # block alpha: -<kl|id> r_kald
                b_dab.contract(
                    "abcd,aced->eb", gooov, out=sigma_s, factor=-1.0
                )
                # 4a)
                # block alpha-beta: <ab|cj>r_ic (jcba,ic)
                eri.contract(
                    "abcd,eb->edac",
                    b_sa,
                    out=sigma_d,
                    factor=1.0,
                    **ovvv,
                )
                # 5)
                # block alpha-beta: -<kb|ij>r_ka
                b_sa.contract("ab,cdae->cbde", gooov, out=sigma_d, factor=-1.0)
            else:
                # 1b)
                # block beta:  f_kc r_iakc
                b_dab.contract("abcd,cd->ab", fia, out=sigma_s, factor=1.0)
                # 2b)
                # block beta:  <ka|cd> r_kcid
                eri.contract(
                    "abcd,aced->eb", b_dab, out=sigma_s, factor=1.0, **ovvv
                )
                # 3b)
                # block beta:  -<lk|ic> r_kcla
                b_dab.contract(
                    "abcd,caeb->ed", gooov, out=sigma_s, factor=-1.0
                )
                # 4b)
                # block alpha-beta: <ab|ic>r_jc (icab)
                eri.contract(
                    "abcd,eb->aced", b_sa, out=sigma_d, factor=1.0, **ovvv
                )
                # 5b)
                # block alpha-beta: -<ka|ji>r_kb
                b_sa.contract("ab,cdae->decb", gooov, out=sigma_d, factor=-1.0)

            sigma.iadd(sigma_s.ravel(), begin0=1 + shift, end0=end_ia + shift)
            sigma.iadd(sigma_d.array.ravel(), begin0=end_jb, end0=end_ab)

        if nacto > 1:
            end_aa = end_ab + nacto * (nacto - 1) * nactv * (nactv - 1) // 4
            for shift, shift2 in zip([0, end_aa - end_ab], [0, end_ia - 1]):
                b_sa.clear()
                sigma_d.clear()
                sigma_s.clear()
                b_daa.clear()
                #
                # Bvectors assignment
                #
                b_sa.assign(bvector, begin2=1 + shift2, end2=end_ia + shift2)
                b_daa.assign(
                    bvector,
                    ind=self.get_index_of_mask(),
                    begin4=end_ab + shift,
                    end4=end_aa + shift,
                )
                tmp = b_daa.copy()
                b_daa.iadd_transpose((2, 1, 0, 3), other=tmp, factor=-1.0)
                b_daa.iadd_transpose((0, 3, 2, 1), other=tmp, factor=-1.0)
                b_daa.iadd_transpose((2, 3, 0, 1), other=tmp, factor=1.0)
                del tmp

                # 1) 1/2f_kc r_iakc
                # block alpha/beta:
                b_daa.contract("abcd,cd->ab", fia, out=sigma_s, factor=0.5)
                # 2) 1/2<la|dc> r_icld
                # block alpha/beta:
                eri.contract(
                    "abcd,edac->eb", b_daa, out=sigma_s, factor=0.5, **ovvv
                )
                # 3) -1/2<la|cd> r_icld
                # block alpha/beta:
                eri.contract(
                    "abcd,ecad->eb",
                    b_daa,
                    out=sigma_s,
                    factor=-0.5,
                    **ovvv,
                )
                # 4) -1/2<kl|id> r_kald
                # block alpha/beta:
                b_daa.contract(
                    "abcd,aced->eb", gooov, out=sigma_s, factor=-0.5
                )
                # 5) 1/2<lk|id> r_kald
                # block alpha/beta:
                b_daa.contract("abcd,caed->eb", gooov, out=sigma_s, factor=0.5)

                # 6) <ab||cj> r_ic
                # block alpha-alpha/beta-beta:
                eri.contract(
                    "abcd,eb->edac", b_sa, out=sigma_d, factor=1.0, **ovvv
                )
                eri.contract(
                    "abcd,eb->ecad", b_sa, out=sigma_d, factor=-1.0, **ovvv
                )
                # 6) -<ab||ci> r_jc
                # block alpha-alpha/beta-beta:
                eri.contract(
                    "abcd,eb->adec", b_sa, out=sigma_d, factor=-1.0, **ovvv
                )
                eri.contract(
                    "abcd,eb->aced", b_sa, out=sigma_d, factor=1.0, **ovvv
                )
                # 7) <ka||ij> r_kb
                # block alpha-alpha/beta-beta:
                b_sa.contract("ab,cdae->cedb", gooov, out=sigma_d, factor=1.0)
                b_sa.contract("ab,cdae->decb", gooov, out=sigma_d, factor=-1.0)
                # 8) -<kb||ij> r_ka
                # block alpha-alpha/beta-beta:
                b_sa.contract("ab,cdae->cbde", gooov, out=sigma_d, factor=-1.0)
                b_sa.contract("ab,cdae->dbce", gooov, out=sigma_d, factor=1.0)

                sigma.iadd(
                    sigma_s.ravel(), begin0=1 + shift2, end0=end_ia + shift2
                )
                sigma.iadd(
                    sigma_d.array[self.get_mask()],
                    begin0=end_ab + shift,
                    end0=end_aa + shift,
                    factor=1.0,
                )
        return sigma

    def build_subspace_hamiltonian_singles(self, bvector, hamiltonian):
        """
        Constructing Hamiltonian subspace for Slater Determinant basis for CIS
        wavefunction

        **Arguments:**

        bvector:
            (OneIndex object) contains current approximation to CIS block
            coefficients in the SD basis.
        """
        self.dimension = "CIS"
        sigma = self.lf.create_one_index(2 * self.dimension - 1)
        b_sa = self.lf.create_one_index(self.dimension)
        b_sa.assign(bvector.array[: self.dimension])
        sigma_a = RCIS.build_subspace_hamiltonian(self, b_sa, hamiltonian)
        sigma.assign(sigma_a, end0=self.dimension)
        b_sa.clear()
        b_sa.set_element(0, bvector.get_element(0))
        b_sa.assign(bvector.array[self.dimension :], begin0=1)
        sigma_a = RCIS.build_subspace_hamiltonian(self, b_sa, hamiltonian)
        sigma.assign(sigma_a.array[1:], begin0=self.dimension)
        self.dimension = "CISD"
        return sigma

    def build_subspace_hamiltonian_doubles(self, bvector):
        """
        Constructing Hamiltonian subspace for Slater Determinant basis for CID
        wavefunction

        **Arguments:**

        bvector:
            (OneIndex object) contains current approximation to CID block
            coefficients in the SD basis.
        """
        self.dimension = "CID"
        sigma = self.lf.create_one_index(self.dimension)
        b_d = self.lf.create_one_index(self.dimension)
        b_d.set_element(0, bvector.get_element(0))
        b_d.assign(bvector.array[2 * self.nacto * self.nactv + 1 :], begin0=1)
        sigma = RCID.build_subspace_hamiltonian_sd(self, b_d)
        self.dimension = "CISD"
        return sigma

    def compute_h_diag(self, *arg):
        """Used by davidson module for pre-conditioning.

        **Returns:**
            (OneIndex object) contains guess vector to davidson diagonalization.

        """
        hdiag = self.lf.create_one_index(self.dimension)
        hdiag_s = self.h_diag_singles()
        hdiag_d = self.h_diag_doubles()
        hdiag.set_element(0, hdiag_s.get_element(0) + hdiag_d.get_element(0))
        hdiag.assign(
            hdiag_s.array[1:], begin0=1, end0=2 * self.nactv * self.nacto + 1
        )
        hdiag.assign(hdiag_d.array[1:], begin0=2 * self.nactv * self.nacto + 1)
        return hdiag

    def h_diag_singles(self):
        """Calculating the guess vector for SD basis.

        hdiag:
            (OneIndex object) contains guess vector for SD basis.
        """
        self.dimension = "CIS"
        hdiag = self.lf.create_one_index(2 * self.dimension - 1)
        hdiag_a = RCIS.compute_h_diag(self)
        hdiag.assign(hdiag_a, end0=self.dimension)
        hdiag.assign(hdiag_a.array[1:], begin0=self.dimension)
        self.dimension = "CISD"
        return hdiag

    def h_diag_doubles(self):
        """Calculating the guess vector for SD basis.

        hdiag:
            (OneIndex object) contains guess vector for SD basis.
        """
        self.dimension = "CID"
        hdiag = RCID.h_diag_sd(self)
        self.dimension = "CISD"
        return hdiag

    def calculate_aux_tensors(self, mo1, mo2):
        """Compute auxiliary matrices.

        **Arguments:**

        mo1, mo2
            One- and two-electron integrals (some Hamiltonian matrix
            elements) in the MO basis.
        """
        self.clear_aux_tensors()
        #
        # Get ranges
        #
        ovvo = self.get_range("ovvo")
        ovov = self.get_range("ovov")
        oooo = self.get_range("oooo")
        ooov = self.get_range("ooov")

        # 1) Fock matrix: fpq
        fock = self.init_aux_tensors("fock", self.nact, self.nact)
        get_fock_matrix(fock, mo1, mo2, self.nacto)

        # 2) temporary matrix: govvo
        govvo = self.init_aux_tensors(
            "govvo", self.nacto, self.nactv, self.nactv, self.nacto
        )
        mo2.slice("abcd->abcd", govvo, **ovvo)

        # 3) temporary matrix: govov
        govov = self.init_aux_tensors(
            "govov", self.nacto, self.nactv, self.nacto, self.nactv
        )
        mo2.slice("abcd->abcd", govov, **ovov)

        # 4) temporary matrix: goooo
        goooo = self.init_aux_tensors(
            "goooo", self.nacto, self.nacto, self.nacto, self.nacto
        )
        mo2.slice("abcd->abcd", goooo, **oooo)

        # 5) temporary matrix: gppqq
        gppqq = self.init_aux_tensors("gppqq", self.nact, self.nact)
        mo2.slice("aabb->ab", gppqq)

        # 6) gooov
        gooov = self.init_aux_tensors(
            "gooov", self.nacto, self.nacto, self.nacto, self.nactv
        )
        mo2.slice("abcd->abcd", gooov, **ooov)

        # 7) temporary matrix: eri = electron repulsion integrals
        if isinstance(self.lf, CholeskyLinalgFactory):
            eri = self.init_aux_tensors(
                "eri",
                self.nact,
                nvec=mo2.nvec,
            )
            eri.assign(mo2)
        else:
            eri = self.init_aux_tensors(
                "eri",
                self.nact,
            )
            eri.assign(mo2)

    def init_aux_tensors(
        self, select, dim0=None, dim1=None, dim2=None, dim3=None, nvec=None
    ):
        # pylint: disable=too-many-arguments
        """Initialize auxiliary matrices.

        **Arguments:**

        select
             (string) label of the auxiliary matrix.

        **Optional Arguments:**

        dim0, dim1, dim2, dim3
             The size of the auxiliary matrix in each dimension. The number of given
             arguments not equal to None determines the order of the matrix.
        """
        check_options(
            "select",
            select,
            "fock",
            "goooo",
            "govov",
            "govvo",
            "eri",
            "gooov",
            "gppqq",
        )

        if select in ["fock", "gppqq"]:
            alloc_size = (self.lf.create_two_index, dim0, dim1)
        elif select in ["govvo", "govov", "goooo", "gooov"]:
            alloc_size = (
                self.denself.create_four_index,
                dim0,
                dim1,
                dim2,
                dim3,
            )
        elif select in ["eri"]:
            alloc_size = (self.lf.create_four_index, dim0, nvec)
        matrix, new = self._cache.load(
            select,
            alloc=alloc_size,
            tags="m",
        )
        if not new:
            raise NonEmptyData(
                f"The matrix {select} already exists. Call clear prior "
                f"to updating the wfn."
            )
        return matrix

    def printer(self):
        """Printing the results."""
        #
        # Local variables
        #
        evals = self.checkpoint["e_ci"]
        evecs = self.checkpoint["civ"]
        threshold = self.threshold
        end_s = 2 * self.nacto * self.nactv + 1
        alpha = "\u03B1"
        beta = "\u03B2"
        data = {
            "spin_block_a",
            "spin_block_b",
            "spin_block_aa",
            "spin_block_bb",
            "spin_block_ab",
            "c_a",
            "c_b",
            "c_ab",
            "c_aa",
            "c_bb",
        }
        #
        # Printing
        #
        log.hline("*")
        log(f"RESULTS OF {self.acronym}")
        log.hline("*")
        log(
            f"{'Root':>5s} {'Exc.Energy[au]':>16s} {'Tot.Energy[au]':>17s}"
            f" {'Weight(s)':>15s}"
            f" {'Weight(d)':>15s}"
        )
        log.hline("_")
        for ind, val in enumerate(evals):
            data = {key: list() for key in data}
            evecsj = evecs[:, ind]
            n_c = np.where(abs(evecsj) > threshold)[0]
            log(
                f"{ind:>3d} {val:> 17.6e} {self.e_ref+val:> 16.8f}"
                f"{np.dot(evecsj[1:end_s], evecsj[1:end_s]):> 17.6f}"
                f"{np.dot(evecsj[end_s:self.dimension], evecsj[end_s:self.dimension]):> 16.6f}"
            )
            log(" ")
            for ind2 in n_c:
                if ind2 == 0:
                    if self.size_consistency_correction:
                        self.rci_corrections.printer()
                    log(f"{'Reference state    C_0:':>31s}")
                    log(f"{evecsj[ind2]:> 34.5f}")
                else:
                    self.collect_data(ind2, data, evecsj)

            if self.csf:
                raise NotImplementedError
            if len(data["spin_block_a"]) > 0:
                log(" ")
                log(f"{'(i->a):':>15s} {'C_ia:':>16s}")
                display(data, "a", alpha)
                display(data, "b", beta)

            if (
                len(data["spin_block_aa"]) > 0
                or len(data["spin_block_ab"]) > 0
            ):
                log(" ")
                log(f"{'(i->a   j->b):':>22s} {'C_iajb:':>11s}")

                display(data, "ab", alpha, beta)
                display(data, "aa", alpha, alpha)
                display(data, "bb", beta, beta)

            log.hline("-")
            log.hline("*")

    def collect_data(self, index, data, evecsj):
        """Collect the data and prepare them for printing:

        **Arguments:**

        *index:
            (int) Number indicating the SD contribution in the CI solution.

        *data:
            (dictionary) Contains two types of data:
            indices (spin_block_[a, b, aa, bb or ab]) of the proper SD spin
            block contributions and the corresponding coefficients
            (c_[a, b, aa, bb or ab]).

        *evecsj:
            Eigenvectors of CI Hamiltonian (without the reference
            state contribution).
        """
        end_a = self.nacto * self.nactv
        end_ab = 2 * end_a + self.nacto * self.nacto * self.nactv * self.nactv
        end_aa = end_ab + ((self.nacto * (self.nacto - 1)) // 2) * (
            (self.nactv * (self.nactv - 1)) // 2
        )

        if index > 2 * end_a:
            i, a, j, b = self.get_index_d(index - 1 - 2 * end_a)
            i, a, j, b = (
                i + self.ncore,
                a + self.ncore,
                j + self.ncore,
                b + self.ncore,
            )
            if index < end_ab:
                data["spin_block_ab"].append([i, a, j, b])
                data["c_ab"].append(evecsj[index])
            elif index < end_aa:
                data["spin_block_aa"].append([i, a, j, b])
                data["c_aa"].append(evecsj[index])
            else:
                data["spin_block_bb"].append([i, a, j, b])
                data["c_bb"].append(evecsj[index])
        elif index > end_a:
            i, a = self.get_index_s(index - 1 - end_a)
            data["spin_block_b"].append(
                [i + self.ncore + 1, a + self.ncore + 1 + self.nacto]
            )
            data["c_b"].append(evecsj[index])
        else:
            i, a = self.get_index_s(index - 1)
            data["spin_block_a"].append(
                [i + self.ncore + 1, a + self.ncore + 1 + self.nacto]
            )
            data["c_a"].append(evecsj[index])

        return data

    def calculate_exact_hamiltonian(self):
        """Calculate the exact hamiltonian of CISD model."""
        raise NotImplementedError
