# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest.linalg.dense import DenseLinalgFactory
from pybest.scf.occ import AufbauOccModel
from pybest.ci.rcid import RCID


testdata_mask = [(6, 28, 0), (6, 28, 1)]


@pytest.mark.parametrize("occ, nbasis, ncore", testdata_mask)
def test_get_mask(occ, nbasis, ncore):
    """Tests if the get_mask function creates the proper shape of
    4-dimensional boolean np.array and checks if the boolean values are assigned
    correctly due to the restrictions on non-redundant and symmetry-unique
    elements of the CI coefficient tensor for double excitations."""
    lf = DenseLinalgFactory(nbasis)
    occ_model = AufbauOccModel(occ)
    rcid = RCID(lf, occ_model, ncore=ncore)

    mask = rcid.get_mask()
    # check shape
    assert mask.shape == (occ - ncore, nbasis - occ, occ - ncore, nbasis - occ)
    # check elements
    for i in range(rcid.nacto):
        for a in range(rcid.nactv):
            for j in range(rcid.nacto):
                for b in range(rcid.nactv):
                    if i < j and a < b:
                        assert mask[i, a, j, b]
                    elif (i >= j) and (a >= b):
                        assert not mask[i, a, j, b]


@pytest.mark.parametrize("occ, nbasis, ncore", testdata_mask)
def test_get_index_of_mask(occ, nbasis, ncore):
    """Checks if the get_index_of_mask function selects the proper indices
    imposed by non-redundant and symmetry-unique elements
    of the CI coefficient tensor for double excitations."""
    lf = DenseLinalgFactory(nbasis)
    occ_model = AufbauOccModel(occ)
    rcid = RCID(lf, occ_model, ncore=ncore)

    mask = rcid.get_index_of_mask()
    dim = len(mask[0])
    for i in range(dim):
        assert mask[0][i] < mask[2][i]
        assert mask[1][i] < mask[3][i]
        assert mask[2][i] < (occ - ncore)
        assert mask[3][i] < (nbasis - occ)


shift = 6 * 6 * 22 * 22
shift1 = 5 * 5 * 22 * 22

testdata_get_index_iajb = [
    (28, 6, 0, shift + 13, (1, 7, 2, 21)),
    (28, 6, 0, shift + 145, (1, 8, 4, 9)),
    (28, 6, 0, shift + 1156, (2, 7, 3, 9)),
    (28, 6, 0, shift + 3333, (5, 12, 6, 17)),
    (28, 6, 1, shift1 + 10, (1, 6, 2, 17)),
    (28, 6, 1, shift1 + 153, (1, 7, 5, 17)),
    (28, 6, 1, shift1 + 1053, (2, 8, 3, 15)),
    (28, 6, 1, shift1 + 2300, (4, 23, 5, 24)),
]


@pytest.mark.parametrize(
    "nbasis, nocc, ncore, ind, expected", testdata_get_index_iajb
)
def test_get_index_iajb(nbasis, nocc, ncore, ind, expected):
    """Checks if the get_index_d function predicts a proper set of active orbital
    indices of some doubly excited SD wavefunction in variants with and without
    frozen core"""
    lf = DenseLinalgFactory(nbasis)
    occ_model = AufbauOccModel(nocc)
    rcid = RCID(lf, occ_model, ncore=ncore)

    i, a, j, b = rcid.get_index_d(ind)
    assert i == expected[0]
    assert a == expected[1]
    assert j == expected[2]
    assert b == expected[3]
