# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file


import numpy as np
import pytest

import pybest.gbasis.gobasis as pybest_basis
from pybest.wrappers import RHF
from pybest.ci.rci_base import RCI
from pybest.ci.rcisd import RCISD
from pybest.ci.test.common import prepare_mol
from pybest.linalg.cholesky import CholeskyLinalgFactory
from pybest.linalg.dense import DenseLinalgFactory
from pybest.scf import AufbauOccModel


# Reference energies
# HF orbitals
e_ex_cisd_hf_ref = [
    -0.204972411018,
    0.216276781554,
    0.222229466166,
    0.288161957788,
]

e_ex_cisd_hf_ref_fc = [
    -0.202959092589,
    0.216578354215,
    0.222528949975,
    0.288486439597,
]

parameters = [
    (4, 0, e_ex_cisd_hf_ref),
    (1, 1, e_ex_cisd_hf_ref_fc),
]


if pybest_basis.PYBEST_CHOLESKY_ENABLED:
    linalg_set = [DenseLinalgFactory, CholeskyLinalgFactory]
else:
    linalg_set = [DenseLinalgFactory]


@pytest.mark.parametrize("nroot, ncore, exp_result", parameters)
@pytest.mark.parametrize("linalg_set", linalg_set)
def test_cisd(nroot, ncore, exp_result, linalg_set):

    kin, na, er, external, olp, lf = prepare_mol(
        "water.xyz", "cc-pvdz", linalg_set
    )

    orb_a = lf.create_orbital()
    occ_model = AufbauOccModel(5)

    hf = RHF(lf, occ_model)
    hf_output = hf(kin, na, er, external, olp, orb_a)

    rcisd = RCISD(lf, occ_model, ncore=ncore)
    rcisd_output = rcisd(kin, na, er, hf_output, nroot=nroot)

    for i in range(nroot):
        assert abs(rcisd_output.e_ci[i] - exp_result[i]) < 1e-6
