#!/usr/bin/env python

from pybest import context
from pybest.gbasis import (
    get_gobasis,
    compute_eri,
    compute_dkh,
    compute_nuclear_repulsion,
    compute_overlap,
)
from pybest.linalg import DenseLinalgFactory
from pybest.scf import AufbauOccModel
from pybest.wrappers import RHF


# Hartree-Fock calculation
# ------------------------

# Load the coordinates from file.
# Use the XYZ file from PyBEST's test data directory.
fn_xyz = context.get_fn("test/water.xyz")

# Create a Gaussian basis set
gobasis = get_gobasis("cc-pvdz", fn_xyz)

# Create a linalg factory
lf = DenseLinalgFactory(gobasis.nbasis)

# Compute integrals
olp = compute_overlap(gobasis)
dkh = compute_dkh(gobasis)
er = compute_eri(gobasis)
external = compute_nuclear_repulsion(gobasis)

# Create alpha orbitals
orb_a = lf.create_orbital()

# Decide how to occupy the orbitals (5 alpha electrons)
occ_model = AufbauOccModel(5)

# Converge RHF
hf = RHF(lf, occ_model)
# the order of the arguments does not matter
hf_output = hf(dkh, er, external, olp, orb_a)

# Write SCF results to a molden file
# First, add gobasis as attribute
hf_output.gobasis = gobasis
# Ready to dump results
hf_output.to_file("water-scf.molden")
