// PyBEST: Pythonic Black-box Electronic Structure Tool
// Copyright (C) 2016-- The PyBEST Development Team
//
// This file is part of PyBEST.
//
// PyBEST is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 3
// of the License, or (at your option) any later version.
//
// PyBEST is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, see <http://www.gnu.org/licenses/>
//
// --

// UPDATELIBDOCTITLE: Gaussian basis set classes

#ifndef PyBEST_BASIS_GBASIS_H
#define PyBEST_BASIS_GBASIS_H

// Libint Gaussian integrals library
#include <libint2.hpp>
#if !LIBINT2_CONSTEXPR_STATICS
#include <libint2/statics_definition.h>
#endif

#include <pybind11/eigen.h>
#include <pybind11/pybind11.h>
namespace py = pybind11;

#if LIBINT2_REALTYPE != double
#error LibInt must be compiled with REALTYPE=double.
#endif

#if LIBINT2_SUPPORT_ERI == 0
#error LibInt must be compiled with support for electron repulsion integrals.
#endif

#ifndef LIBINT2_MAX_AM
#define LIBINT2_MAX_AM LIBINT2_MAX_AM_ERI
#endif
#if LIBINT2_MAX_AM < MAX_SHELL_TYPE
#error LibInt must be compiled with an angular momentum limit of at least MAX_SHELL_TYPE.
#endif

#include <libint2/lcao/molden.h>

#include <vector>

class Basis {
 public:
  const std::string basisname;
  const std::string molfile;
  const std::string dirname;
  std::size_t nbasis;
  std::size_t nshell;
  std::size_t ncenter;
  // we need to store this information to ensure the functionality of
  // some pybest modules
  std::vector<long> nprim;
  std::vector<double> alpha;
  std::vector<double> contraction;
  std::vector<long> shell2atom;
  std::vector<int> shell_types;
  std::vector<long> atomic_numbers;
  std::vector<std::vector<double>> coordinates;
  Basis(const std::string &basisname, const std::string &molfile,
        const std::string &dirname);
  Basis(const std::string &molfile, std::vector<long> nprims,
        std::vector<long> center, std::vector<int> l,
        std::vector<double> alphas, std::vector<double> contr);
  // used in from_coordinates() which is exported to python
  Basis(std::vector<libint2::Atom> atoms_, std::vector<long> nprims,
        std::vector<long> center, std::vector<int> l,
        std::vector<double> alphas, std::vector<double> contr);
  Basis(const Basis &rhs) = default;
  ~Basis() = default;

  // Members:
  libint2::BasisSet obs;
  std::vector<libint2::Atom> atoms;

  // Getters:
  const std::string &getBasisname() const;
  const std::string &getMolfile() const;
  const std::string &getDirname() const;
  std::size_t getNBasis();
  std::size_t getNShell();
  std::size_t getNCenter();
  std::vector<long> getNPrim();
  std::vector<double> getAlpha();
  std::vector<double> getContraction();
  std::vector<long> getShell2atom();
  std::vector<int> getShellTypes();
  std::vector<long> getAtomicNumbers();
  std::vector<std::vector<double>> getCoordinates();

  // More general functions accessible through python
  void print_basis_info();
  void print_atomic_info();
  std::size_t get_nbasis_in_shell(std::size_t s);
  void renormalize_contr(int l, std::vector<std::size_t> n, double factor);
  double get_renorm(std::size_t l, std::size_t s, std::size_t n);
  double get_renorm(std::size_t l, std::size_t s, std::size_t na,
                    std::vector<std::size_t> n);
  void dump_molden(const char *filename,
                   const py::EigenDRef<Eigen::MatrixXd> coeffs,
                   const py::EigenDRef<Eigen::MatrixXd> occs,
                   const py::EigenDRef<Eigen::MatrixXd> energies,
                   const std::vector<bool> &spin, double eps);
  // all remaining functions only visible to C++ code
  void pure2cart(int l);
  void set_dummy_atom(int i);
  std::size_t get_shell_size(int shell_type);
  void assign_basis_set(std::vector<long> nprims, std::vector<long> center,
                        std::vector<int> l, std::vector<double> alphas,
                        std::vector<double> contr);
  void to_hdf5(const std::string &filename);
};

Basis from_coordinates(std::vector<long> atomic_numbers,
                       std::vector<std::vector<double>> coordinates,
                       std::vector<long> primitives,
                       std::vector<long> shell2atom,
                       std::vector<int> shell_types, std::vector<double> alphas,
                       std::vector<double> contraction);
#endif
