# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: skip-file

import numpy as np
import pytest

from pybest.linalg.dense import DenseLinalgFactory
from pybest.modelhamiltonians.physmodham import Hubbard
from pybest.wrappers.hf import RHF
from pybest.scf import AufbauOccModel
from pybest.geminals.rpccd import ROOpCCD
from pybest.cc import RpCCDLCCD, RpCCDLCCSD
from pybest.orbital_entanglement.orbital_entanglement import (
    OrbitalEntanglementRpCCDLCC,
)

s1_ref = np.array(
    [
        0.01645669276751,
        0.01963600360018,
        0.01965325872394,
        0.01965325843777,
        0.01963600411211,
        0.01645669255274,
    ]
)

mi_ref = np.array(
    [
        [
            [0.0, 0.00613942, 0.00614475, 0.00692526, 0.00691559, 0.01308797],
            [0.00613942, 0.0, 0.00360408, 0.00357657, 0.02326247, 0.00691559],
            [0.00614475, 0.00360408, 0.0, 0.02327683, 0.00357657, 0.00692526],
            [0.00692526, 0.00357657, 0.02327683, 0.0, 0.00360408, 0.00614475],
            [0.00691559, 0.02326247, 0.00357657, 0.00360408, 0.0, 0.00613942],
            [0.01308797, 0.00691559, 0.00692526, 0.00614475, 0.00613942, 0.0],
        ],
    ]
)


def prepare_ham(lf, t, v):
    modelham = Hubbard(lf, pbc=True)

    hopping = modelham.compute_one_body(t)
    onsite = modelham.compute_two_body(v)

    olp = modelham.compute_overlap()

    return hopping, onsite, olp


def test_pccd_lcc():
    # Test only for Hubbard
    lf = DenseLinalgFactory(6)
    one, two, olp = prepare_ham(lf, -1.0, 0.5)

    occ_model = AufbauOccModel(3)
    orb_a = lf.create_orbital(6)

    hf = RHF(lf, occ_model)
    hf_ = hf(one, two, orb_a, olp)

    pccd = ROOpCCD(lf, occ_model)
    pccd_ = pccd(one, two, hf_)

    lccsd = RpCCDLCCSD(lf, occ_model)
    lccsd_ = lccsd(one, two, pccd_, l=True)

    oe = OrbitalEntanglementRpCCDLCC(lf, lccsd_)
    oe_ = oe()

    assert np.allclose(s1_ref, oe_.s_1._array, atol=1e-4)
    assert np.allclose(mi_ref, oe_.I_12._array, atol=1e-4)


# Dry run:
def test_pccd_core():
    # Test only for Hubbard
    # This test makes no sense as no orbitals should be frozen
    lf = DenseLinalgFactory(6)
    one, two, olp = prepare_ham(lf, -1.0, 0.5)

    occ_model = AufbauOccModel(3)
    orb_a = lf.create_orbital(6)

    hf = RHF(lf, occ_model)
    hf_ = hf(one, two, orb_a, olp)

    pccd = ROOpCCD(lf, occ_model, ncore=1)
    pccd_ = pccd(one, two, hf_)

    lccsd = RpCCDLCCSD(lf, occ_model, ncore=1)
    lccsd_ = lccsd(one, two, pccd_, l=True)

    oe = OrbitalEntanglementRpCCDLCC(lf, lccsd_)
    oe_ = oe()
